import copy


class Parameter:

    type_mapping = {
        "float": (int, float),
        "integer": int,
        "bool": bool,
        "list": list,
        "dict": dict,
        "string": str,
        "enum": object,
    }

    def __init__(self, param: dict, name):
        self.name = name
        self.description = param["description"]
        self.type = self.type_mapping[param["type"]]
        self.enum = param.get("enum")
        self.optional = param.get("optional", False)
        self.value = param["default"]

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if self.optional and val is None:
            self._value = val
            return
        if not isinstance(val, self.type):
            raise ValueError(
                f"Setting parameter {self.name} of type {self.type} with value {val}."
            )
        elif self.enum and val not in self.enum:
            raise ValueError(
                f"Setting parameter {self.name} with a value {val} that is not in {self.enum}."
            )
        self._value = val


class FGParameters(object):
    def __init__(self, parameters: dict):
        self.parameter = {}
        for name, spec in parameters.items():
            self.parameter[name] = Parameter(spec, name)

    def __getitem__(self, key):
        return self.parameter[key].value

    def keys(self):
        return self.parameter.keys()

    def update(self, values):
        for name, value in values.items():
            if name not in self.parameter:
                raise KeyError(f"Parameter {name} not found in manifest.json.")
            self.parameter[name].value = value

    def copy(self):
        return copy.deepcopy(self)

    def check(self, name, checker, error_message):
        if not checker(self[name]):
            raise ValueError(f'Parameter "{name}" failed validation: {error_message}')



import logging
import jsonschema
import pytest

from fastgenomics.app import FGApp
from fastgenomics.process import FGProcess


def test_fgapp(app_dir):
    FGApp(app_dir)


def test_app_validation_error(app_dir_validation_error):
    with pytest.raises(jsonschema.ValidationError):
        FGApp(app_dir_validation_error)


def test_app_runtime_error(app_dir_runtime_error):
    with pytest.raises(RuntimeError):
        FGApp(app_dir_runtime_error)


def test_app_file_not_found(app_dir_missing_files):
    with pytest.raises(FileNotFoundError) as e:
        FGApp(app_dir_missing_files)
    assert "LICENSE" in str(e.value)
    assert "LICENSE-THIRD-PARTY" in str(e.value)
    assert "manifest.json" in str(e.value)


def test_app_non_existing(app_dir_non_existing):
    with pytest.raises(FileNotFoundError) as e:
        FGApp(app_dir_non_existing)
    assert "Could not find the App directory" in str(e.value)


def test_fgprocess_1(app_dir, data_root, caplog):
    caplog.set_level(logging.DEBUG)
    FGProcess(app_dir, data_root)
    assert any(
        ["Parameters" in x.message for x in caplog.records]
    ), "Parameter logs are not set"


def test_fgprocess_2(app_dir, data_root_2, caplog):
    caplog.set_level(logging.DEBUG)
    FGProcess(app_dir, data_root_2)
    assert any(
        ["Parameters" in x.message for x in caplog.records]
    ), "Parameter logs are not set"



# Flit documentation build configuration file
# http://www.sphinx-doc.org/en/stable/config.html

import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent))
from fastgenomics import __version__


needs_sphinx = "1.3"

extensions = [
    "sphinx.ext.autodoc",  # API docs
    "sphinx.ext.napoleon",  # google-style docstrings
    "sphinx.ext.intersphinx",  # links to official docs
    "sphinx_autodoc_typehints",
]

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    jsonschema=("https://python-jsonschema.readthedocs.io/en/v2.6.0/", None),
)

templates_path = ["_templates"]
exclude_patterns = ["_build"]
html_static_path = ["_static"]

source_suffix = ".rst"
master_doc = "index"

project = u"FASTGenomics"
copyright = u"2018, Comma Soft"

release = __version__
version = __version__  # short version

pygments_style = "sphinx"
html_theme = "sphinx_rtd_theme"
# html_theme_options = {}

# html_logo = '_static/flit_logo_nobg_cropped.svg'
# html_favicon = None

html_context = {
    "display_github": True,
    "github_user": "fastgenomics",
    "github_repo": "fastgenomics-py",
    "github_version": "master",
    "conf_py_path": "/doc/",
}



import pytest
from jinja2 import UndefinedError


def test_can_write_summary(summary):
    summary.write(test=1)


def test_missing_values_in_templates_raise(summary):
    with pytest.raises(UndefinedError):
        summary.write()


def test_summary_includes_parameters(summary, fgprocess):
    summary.write(test=1)
    text = summary.output.read_text()

    assert "### Parameters" in text

    # check the str_value
    assert "str_value" in text
    assert "hello from parameters.json" in text
    assert "The column denoting the batches" in text

    # check the int_value
    assert "int_value" in text
    assert "150" in text
    assert "Number of Dimensions" in text



import pytest
from pathlib import Path
from fastgenomics.process import FGProcess
from fastgenomics.testing import cleanoutput

HERE = Path(__file__).parent
DATA = HERE / "data"
APPS = HERE / "apps"


@pytest.fixture
def app_dir():
    return APPS / "working"


@pytest.fixture
def app_dir_validation_error():
    return APPS / "validation_error"


@pytest.fixture
def app_dir_missing_files():
    return APPS / "missing_files"


@pytest.fixture
def app_dir_non_existing():
    path = APPS / "non_existing"
    assert not path.exists()
    return path


@pytest.fixture
def app_dir_runtime_error():
    return APPS / "runtime_error"


@pytest.fixture
def data_root():
    return DATA / "working_1"


@pytest.fixture
def data_root_2():
    return DATA / "working_2"


@pytest.fixture
def data_root_none():
    return DATA / "non_existing"


@pytest.fixture
def data_root_missing_files():
    return DATA / "missing_files"


@pytest.fixture
def fgprocess(app_dir, data_root):
    fg = FGProcess(app_dir, data_root)
    cleanoutput(fg)
    return fg


@pytest.fixture
def summary(fgprocess):
    fgprocess.summary.template = fgprocess.app.app_dir / "summary.md.j2"
    return fgprocess.summary


@pytest.mark.anndata
@pytest.fixture
def adata(fgprocess):
    import fastgenomics.external.anndata as fgad

    return fgad.read_data(
        fgprocess, expr="some_input", gene_meta="genes", cell_meta="cells"
    )



import os
from pathlib import Path

# set default paths
DEFAULT_APP_DIR = Path(os.environ.get("FG_APP_DIR", "/app"))
DEFAULT_DATA_ROOT = Path(os.environ.get("FG_DATA_ROOT", "/fastgenomics"))

# summary key for the FGProcess.output
SUMMARY_KEY = "summary"



import fastgenomics.external.anndata as fgad
import pytest
import pandas as pd


@pytest.mark.anndata
def test_read(fgprocess):
    adata = fgad.read_data(
        fgprocess, expr="some_input", gene_meta="genes", cell_meta="cells"
    )

    assert (adata.obs.index == ["c1", "c3", "c4", "c2"]).all()
    assert (adata.var.index == [1, 3, 5, 10, 100]).all()
    assert adata.X[1, 0] == 4
    assert adata.shape == (4, 5)
    assert adata.var.shape[1] == 2
    assert adata.obs.shape[1] == 2
    assert "some_column" in adata.var
    assert "some_column" in adata.obs
    assert "batch_id" in adata.obs


@pytest.mark.anndata
def test_read_type_throws(fgprocess):
    with pytest.raises(TypeError):
        fgprocess.input["some_input"].type = "wrong_type"
        fgad.read_data(
            fgprocess, expr="some_input", gene_meta="genes", cell_meta="cells"
        )


@pytest.mark.anndata
def test_read_output_throws(fgprocess):
    with pytest.raises(KeyError):
        fgad.read_data(
            fgprocess, expr="some_output", gene_meta="genes", cell_meta="cells"
        )


@pytest.mark.anndata
def test_write(fgprocess, adata):
    for f in ["count_output", "genes_output", "cells_output"]:
        assert not fgprocess.output[f].path.exists()

    fgad.write_data(
        fgprocess,
        adata,
        expr="count_output",
        gene_meta="genes_output",
        cell_meta="cells_output",
    )

    for f in ["count_output", "genes_output", "cells_output"]:
        assert fgprocess.output[f].path.exists()


@pytest.mark.anndata
def test_read_write(fgprocess, adata):
    fgad.write_data(
        fgprocess,
        adata,
        expr="count_output",
        gene_meta="genes_output",
        cell_meta="cells_output",
    )

    expr_out = pd.read_csv(fgprocess.output["count_output"].path)
    expr_out = expr_out.sort_values(list(expr_out.columns)).reset_index(drop=True)

    genes_out = pd.read_csv(fgprocess.output["genes_output"].path)
    genes_out.set_index("entrez_id", inplace=True)

    cells_out = pd.read_csv(fgprocess.output["cells_output"].path)
    cells_out.set_index("cell_id", inplace=True)

    expr_in = pd.DataFrame(
        dict(
            cell_id=["c1", "c1", "c2", "c3", "c3"],
            entrez_id=[1, 3, 5, 1, 100],
            expression=[1.0, 2.0, 3.0, 4.0, 1.0],
        )
    )
    expr_in = expr_in.sort_values(list(expr_out.columns)).reset_index(drop=True)

    assert expr_out.equals(expr_in)
    assert genes_out.equals(adata.var)
    assert cells_out.equals(adata.obs)



from pytest import raises


def test_checker_true(fgprocess):
    fgprocess.parameters.check("str_value", lambda x: True, "All good!")
    fgprocess.parameters.check("int_value", lambda x: x == 150, "All good!")


def test_cheker_raises(fgprocess):
    with raises(ValueError, match=r".*Failed!"):
        fgprocess.parameters.check("str_value", lambda x: False, "Failed!")

    with raises(ValueError, match=r".*Number of dimensions.*"):
        fgprocess.parameters.check(
            "int_value",
            lambda x: x > 150,
            "Number of dimensions has to be greater then 150",
        )



from fastgenomics.process import FGProcess
import pytest


def test_io_throws_on_non_existing_files(app_dir, data_root_none):
    with pytest.raises(FileNotFoundError):
        FGProcess(app_dir, data_root_none)



from fastgenomics.data import FGData
import pytest


def test_data_1(data_root):
    FGData(data_root)


def test_data_2(data_root_2):
    FGData(data_root_2)


def test_data_file_not_found(data_root_missing_files):
    with pytest.raises(FileNotFoundError) as e:
        FGData(data_root_missing_files)

    files = ["data", "config", "output", "summary", "config/input_file_mapping.json"]
    for file in files:
        assert file in str(e.value)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastgenomics', 'fastgenomics.external']

package_data = \
{'': ['*'], 'fastgenomics': ['schemes/*', 'templates/*']}

install_requires = \
['deprecated~=1.2.4',
 'jsonschema~=2.6',
 'Jinja2~=2.9',
 'get_version~=2.0.1',
 'setuptools']

extras_require = \
{'doc': ['sphinx', 'sphinx_rtd_theme', 'sphinx-autodoc-typehints'],
 'test': ['pytest >=4', 'anndata >=0.6']}

setup(name='fastgenomics',
      version='2.0.1',
      description='FASTGenomics python helper',
      author='FASTGenomics Team',
      author_email='contact@fastgenomics.org',
      url='https://github.com/fastgenomics/fastgenomics-py',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.6',
     )



from pathlib import Path
from collections import OrderedDict
import json

from .defaults import DEFAULT_DATA_ROOT

DATA_SUBDIRS = ["data", "config", "output", "summary"]


class FGData(object):
    """This class stores the paths to data structured according to the
fastgenomics specification.  It also loads the input file mappings and
checks if the files exist.

    """

    _subdirs = DATA_SUBDIRS
    _mandatory_files = ["config/input_file_mapping.json"] + DATA_SUBDIRS

    def __init__(self, data_root=DEFAULT_DATA_ROOT):
        if isinstance(data_root, str):
            data_root = Path(data_root)
        self.root = data_root
        self.check_files()

        self.paths = self.get_paths()
        self.input_file_mapping = self.get_input_file_mapping()
        self.parameters = self.get_parameters()

    def get_input_file_mapping(self):
        mapping_file = self.paths["config"] / "input_file_mapping.json"
        return json.loads(mapping_file.read_bytes())

    def get_paths(self):
        return {dir: self.root / dir for dir in self._subdirs}

    def get_parameters(self):
        params_file = self.paths["config"] / "parameters.json"
        if params_file.exists():
            return json.loads(params_file.read_text())
        else:
            return {}

    def check_files(self):
        if not self.root.exists():
            raise FileNotFoundError(
                f"Could not find the data directory under {self.root}"
            )

        not_found = []
        for f in self._mandatory_files:
            absolute_path = self.root / f
            if not absolute_path.exists():
                not_found.append(f)

        if not_found:
            msg = [
                f'Could not find the following files or directories in "{self.root}":'
            ]
            msg += [f"- {file}" for file in not_found]
            raise FileNotFoundError("\n".join(msg))



try:
    import pandas as pd
    import numpy as np
    import scipy.sparse as sp
    from anndata import AnnData
except ImportError as e:
    msg = f"Could not import some of the necessary modules ({e.name}).  Please make sure to install anndata (https://github.com/theislab/anndata) with all its dependencies correctly (e.g. pandas, numpy, scipy)."
    raise ImportError(msg, name=e.name, path=e.path)
    raise e


def get_path(file, content_type):
    """Returns a path if the file is of requested FASTGenomics type,
otherwise throws a TypeError."""
    if file is None:
        return None
    if file.type != content_type:
        raise TypeError(
            f'File "{file.name}" is of type "{file.type}" but expected "{content_type}".'
        )
    return file.path


def read_data(
    fgprocess,
    expr="expression_matrix",
    cell_meta="cell_metadata",
    gene_meta="gene_metadata",
):
    """Reads an anndata object by composing three files together:
`expression_matrix`, `cell_metadata` and `gene_metadata`."""
    expr_path = get_path(fgprocess.input[expr], content_type="expression_matrix")
    expr = pd.read_csv(expr_path)

    obs = read_cell_metadata(fgprocess, cell_meta, expr)
    var = read_gene_metadata(fgprocess, gene_meta, expr)

    counts = read_sparse_matrix(expr, obs, var)

    adata = AnnData(counts, obs=obs, var=var, dtype="float64")
    return adata


def read_cell_metadata(fgprocess, cell_meta, expr):
    if cell_meta is None:
        cell_path = None
    else:
        cell_path = get_path(fgprocess.input[cell_meta], content_type="cell_metadata")

    expr_cell_id = expr.cell_id.unique()

    if cell_path is None:
        df = pd.DataFrame(expr_cell_id, columns=["cell_id"]).set_index("cell_id")
    else:
        df = pd.read_csv(cell_path, index_col="cell_id")
        if set(expr_cell_id) - set(df.index):
            raise Exception(
                f'Some cell_id\'s were present in the expression matrix but not in "{cell_meta}".'
            )
    return df


def read_gene_metadata(fgprocess, gene_meta, expr):
    gene_path = get_path(fgprocess.input[gene_meta], content_type="gene_metadata")

    expr_entrez_id = expr.entrez_id.unique()

    if gene_path is None:
        df = pd.DataFrame(expr_entrez_id, columns=["entrez_id"]).set_index("entrez_id")
    else:
        df = pd.read_csv(gene_path, index_col="entrez_id")
        if set(expr_entrez_id) - set(df.index):
            raise Exception(
                f'Some entrez_id\'s were present in the expression matrix but not in "{gene_meta}".'
            )
    return df


def read_sparse_matrix(expr, obs, var):
    cell_idx = pd.DataFrame(
        dict(cell_id=obs.index, cell_idx=np.arange(obs.shape[0]))
    ).set_index("cell_id")
    entrez_idx = pd.DataFrame(
        dict(entrez_id=var.index, entrez_idx=np.arange(var.shape[0]))
    ).set_index("entrez_id")
    expr = expr.merge(cell_idx, on="cell_id", copy=False)
    expr = expr.merge(entrez_idx, on="entrez_id", copy=False)

    counts = sp.coo_matrix(
        (expr.expression, (expr.cell_idx, expr.entrez_idx)),
        shape=(obs.shape[0], var.shape[0]),
    ).tocsr()

    return counts


# Writing
def write_exprs_csv(adata, csv_file):
    mat = adata.X.tocoo()
    df = pd.DataFrame.from_dict(
        dict(
            cell_id=adata.obs_names[mat.row],
            entrez_id=adata.var_names[mat.col],
            expression=mat.data,
        )
    )
    df.to_csv(csv_file, index=False)


def write_data(fgprocess, adata, expr=None, cell_meta=None, gene_meta=None):
    if expr is not None:
        exprs_path = get_path(fgprocess.output[expr], content_type="expression_matrix")
        write_exprs_csv(adata, exprs_path)
    if cell_meta is not None:
        cell_path = get_path(fgprocess.output[cell_meta], content_type="cell_metadata")
        adata.obs.to_csv(cell_path)
    if gene_meta is not None:
        gene_path = get_path(fgprocess.output[gene_meta], content_type="gene_metadata")
        adata.var.to_csv(gene_path)
